/*
 * $Id: arch-solaris-threads.c,v 1.4 1997/06/09 19:03:05 athena Exp $
 */
/*
 * $Log: arch-solaris-threads.c,v $
 * Revision 1.4  1997/06/09 19:03:05  athena
 * Changed RTS so that Cilk won't dump core by default
 *
 * Revision 1.3  1997/03/02 21:06:25  randall
 * Made sure Cilk_die doesn't return.
 *
 * Revision 1.2  1997/03/01 17:42:31  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/tspriocntl.h>
#include <sys/procfs.h>

DECLARE_SHARED_RO_VARIABLE(static int, lock_type, );
#define MEMORY_LOCKS 0
#define OS_LOCKS 1

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

DECLARE_SHARED_VARIABLE(static Cilk_mutex, dprintf_lock,);

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: ", Self);
     else
	  fprintf(stderr, "global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));
}

DECLARE_SHARED_VARIABLE(static Cilk_mutex, die_lock,);

void Cilk_die(const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: fatal error: ", Self);
     else
	  fprintf(stderr, "global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (USE_SHARED_RO(dump_core) && Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
     else
          exit(0);              /* don't dump core */
}

/***********************************************************\
 * Mapping of variable segments
\***********************************************************/

#if !CILK_MANAGE_SHARED_SEGMENT
/*
 * if linker segments are not used, a portable version
 * of these two functions is provided in cilk.c
 */

/* 
 * map the global variables segment as shared
 */
void Cilk_map_shared_segment(void)
{
     /* nothing needs to be done */
}

#endif

#if !CILK_MANAGE_PRIVATE_SEGMENT
void Cilk_map_private_segment(void)
{
     /* variables are private by default */
}

#endif

/***********************************************************\
 * Malloc support
\***********************************************************/
/*
 * we could use malloc() instead of mmap(), since under
 * Solaris threads malloc() gives back shared memory anyway.
 * However, in case of core dump, the whole heap is dumped
 * (64Mb?), which is too much.
 */
void *Cilk_map_shared(size_t len)
{
     int fd;
     void *area;

     fd = open("/dev/zero", O_RDWR);

     CILK_CHECK(fd >= 0, ("Open error in Cilk_map_shared\n"));

     area = mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
     if (area == MAP_FAILED) {
	  /* 
	   * can't use Cilk_die, because the mutex in
	   * not initialized yet
	   */
	  fprintf(stderr, "mmap error in Cilk_map_shared -- "
		   "couldn't allocate %d bytes\n", len);
	  exit(1);
     }
     close(fd);
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
     USE_SHARED_RO(lock_type) = MEMORY_LOCKS;

     for (; i < argc; ++i) {
	  if (*argv[i] != '-')
	       break;

	  if (strcmp(argv[i] + 1, "-") == 0) {
	       ++i;
	       break;
	  } else if (strcmp(argv[i] + 1, "os-locks") == 0) {
	       USE_SHARED_RO(lock_type) = OS_LOCKS;
	  } else if (strcmp(argv[i] + 1, "memory-locks") == 0) {
	       USE_SHARED_RO(lock_type) = MEMORY_LOCKS;
	  } else
	       break;
     }

     return i;
}

/***********************************************************\
 * Support for mutexen 
\***********************************************************/

#if SPARC_V9
#define ACQUIRE(lock) Cilk_cas(lock, 0, 1)
#define RELEASE(lock) *lock = 0
#endif

#if SPARC_V8
#define ACQUIRE(lock) Cilk_swap(lock, 1)
#define RELEASE(lock) Cilk_swap(lock, 0)
#endif

/*******************************
 * Memory locks
 *******************************/
/* initialize lock */
static inline void Cilk_mutex_init_memory(Cilk_mutex *lock)
{
     lock->memory = 1;
     RELEASE(lock);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_memory(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     /* 
      * This hackery should reduce bus traffic, if the
      * cache is snooped correctly.
      */
     while (ACQUIRE(&lock->memory) != 0) {
	  while (lock->memory != 0);
     }

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_memory(Cilk_mutex *lock)
{
     RELEASE(&lock->memory);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_memory(Cilk_mutex *lock)
{
     if (ACQUIRE(&lock->memory) == 0) {
	  return 1;
     } else
	  return 0;
}

/*******************************
 * Operating System locks
 *******************************/
/* initialize lock */
static inline void Cilk_mutex_init_os(Cilk_mutex *lock)
{
     mutex_init(&lock->os, USYNC_PROCESS, NULL);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_os(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     mutex_lock(&lock->os);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_os(Cilk_mutex *lock)
{
     mutex_unlock(&lock->os);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_os(Cilk_mutex *lock)
{
     if (mutex_trylock(&lock->os) == 0) {
	  return 1;
     } else {
	  return 0;
     }
}

/*********************
 * Lock glue 
 *********************/
#define MAKE_LOCKFUN(name) \
void name(Cilk_mutex *lock)                                   \
{                                                             \
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {	      \
	  name##_memory(lock);				      \
     } else {						      \
	  name##_os(lock);				      \
     }							      \
}							      \
							       
MAKE_LOCKFUN(Cilk_mutex_init)
MAKE_LOCKFUN(Cilk_mutex_wait)
MAKE_LOCKFUN(Cilk_mutex_signal)

int Cilk_mutex_try(Cilk_mutex *lock)
{
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {
	  return Cilk_mutex_try_memory(lock);
     } else {
	  return Cilk_mutex_try_os(lock);
     }
}

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>
#include <sys/resource.h>

/*
 * Start forking processes: one process enters this function,
 * <Cilk_active_size> processes should exit this function immediately
 * and return 0.  Each returning process should set <Self> to a unique
 * number between 0 and <Cilk_active_size> - 1.
 */
void Cilk_create_children(void (*child) (int))
{
     int i;
     int res;
     void *status;
     thread_t *tid;

     tid = malloc(USE_SHARED_RO(active_size) * sizeof(thread_t));

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	  res = thr_create(NULL, 0,
			   (void * (*) (void *))child,
			   (void *) i, THR_BOUND, tid + i);
	  if (res) {
	       Cilk_die("Can't create threads\n");
	  }
     }

     while (thr_join(0 , NULL, &status) == 0);
}

/***************************************************
 * Timer control
 ***************************************************/
static int openslashproc(int pid)
{
     char buf[64];
     int fd;

     sprintf(buf, "/proc/%d", pid);
     if ((fd = open(buf, 2)) < 0)
	  perror("open /proc");

     return fd;
}

static void start_microaccounting(void)
{
     int fd, flags;

     fd = openslashproc(getpid());

     flags = PR_MSACCT;
     if (ioctl(fd, PIOCSET, &flags, sizeof(flags)) < 0) {
	  perror("ioctl set microstate acct");
     }
}

/**********************************************************
 * Priority control
 **********************************************************/
#define LOW_PRIORITY (-60)

DECLARE_PRIVATE_VARIABLE(static int, current_priority,);
void Cilk_lower_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) != LOW_PRIORITY) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_LWPID, _lwp_self(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = tsparams->ts_upri = LOW_PRIORITY;
	  priocntl(P_LWPID, _lwp_self(), PC_SETPARMS, (caddr_t) & params);
     }
}

void Cilk_raise_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) == LOW_PRIORITY) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_LWPID, _lwp_self(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = 
	       tsparams->ts_upri = tsparams->ts_uprilim;
	  priocntl(P_LWPID, _lwp_self(), PC_SETPARMS, (caddr_t) & params);
     }
}

/**********************************************************
 * Initialization of the module
 **********************************************************/

void Cilk_arch_specific_init(void)
{
     Cilk_mutex_init(&USE_SHARED(dprintf_lock));
     Cilk_mutex_init(&USE_SHARED(die_lock));
     start_microaccounting();
}

void Cilk_arch_specific_per_process_init(void)
{
     USE_PRIVATE(current_priority) = -1;
     start_microaccounting();
}
