/*
 * $Id: cilk.h,v 1.3 1997/03/01 17:42:47 athena Exp $
 */
/*
 * $Log: cilk.h,v $
 * Revision 1.3  1997/03/01 17:42:47  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_H
#define _CILK_H

#include <cilk-conf.h>

/***********************************************************\
 * Architecture-specific definitions and functions
\***********************************************************/

#include CILK_ARCH_H

/***********************************************************\
 * Types used by cilk2c output
\***********************************************************/

/*
 * ProcInfo fields:
 *          |       NORMAL           |               INLET
 * ---------+------------------------+---------------------------------------
 *     size | size of return value   | size of returned (by SPAWN) value
 *    index | index to put ret. val. | index to put ret. (by INLET_CALL) val.
 *    inlet | 0                      | pointer to inlet function
 *  argsize | 0                      | size of inlet's arguments in frame
 * argindex | ???                    | index of inlet's arguments in frame
 *
 * Hack: the field "inlet" for entry 0 contains the address of 
 * the slow function.  The field "size" for entry 0 contains the size
 * of the value returned by the function as a whole (as opposed to the
 * other entries, which contain the size of the value returned by that
 * spawn/sync statement).
 */

typedef struct {
     int size;
     int index;
     void (*inlet) ();
     int argsize;
     int argindex;
} ProcInfo;

/*
 * a linked list of blocks allocated by Cilk_alloca.
 */
struct alloca_header {
     struct alloca_header *next;
     size_t size;
};

/* 
 * a stack frame consists of this header and of a procedure-specific
 * part 
 */
typedef struct StackFrame {
     WHEN_CILK_DEBUG(volatile int magic;)
     int entry;
     int index;			/* 
				 * index outstanding spawn will return to, 
				 * valid only if entry does not specify
				 * it. 
				 */
     ProcInfo *sig;
     WHEN_CILK_ALLOCA(struct alloca_header *alloca_h;)
     WHEN_CILK_TIMING(Cilk_time mycp;)
     WHEN_CILK_TIMING(Cilk_time cp;)
     WHEN_CILK_TIMING(Cilk_time work;)
} StackFrame;

/*
 * a stack is an array of frame pointers 
 */
typedef StackFrame **Stack;

/*
 * The way head and tail work is a bit tricky.  Conceptually,
 * head and tail belong to Closure; however, when a steal
 * occurs, the Closure of a certain queue is replaced by
 * another Closure.  There is no way to initialize the
 * new Closure with the appropriate head and tail, and install
 * it atomically, unless we agree to lock the victim processor
 * (and we don't want to).  Therefore, the closure contains
 * *pointers* to head and tail.  The actual head and tail are allocated
 * on a per-processor basis.  
 *
 * We allocate this little structure to keep the values of head
 * and tail.  It also caches other important quantities used by
 * the cilk2c output, that should reside in the Closure but
 * are cached here for efficiency reasons.
 */
typedef struct ClosureCache {
     volatile StackFrame **head, **tail;
     /* see sched.c for a description of the exception-handling machinery */
     volatile StackFrame **exception;  
     Stack stack;
     CILK_CACHE_LINE_PAD;
} ClosureCache;

/*
 * Functions defined in the architecture-specific file and used
 * in Cilk programs
 */
extern void Cilk_mutex_init(Cilk_mutex *lock);
extern void Cilk_mutex_wait(Cilk_mutex *lock);
extern void Cilk_mutex_signal(Cilk_mutex *lock);
extern int Cilk_mutex_try(Cilk_mutex *lock);
extern void Cilk_dprintf(const char *fmt,...);
extern void Cilk_die(const char *fmt,...);

/*
 * Scheduler functions used in Cilk programs
 */
extern int Cilk_sync(void);
extern int Cilk_exception_handler(void *, int);
extern void Cilk_set_result(void *resultp, int size);
extern void *Cilk_internal_malloc(size_t);
extern void Cilk_internal_free(void *p, size_t size);
extern void Cilk_internal_malloc_final_check(void);
extern void Cilk_after_sync_slow_cp(Cilk_time *work, Cilk_time *cp);
extern void Cilk_abort_standalone(void);
extern void Cilk_abort_slow(void);
extern void Cilk_event_new_thread(void);

/* variables */
DECLARE_PRIVATE_VARIABLE(extern int, _Self, );
#define Self USE_PRIVATE(_Self)

DECLARE_SHARED_RO_VARIABLE(extern size_t, stackdepth, );

DECLARE_SHARED_RO_VARIABLE(extern const char *, assertion_failed_msg ,);

DECLARE_SHARED_VARIABLE(extern Cilk_time, critical_path, );
DECLARE_SHARED_VARIABLE(extern Cilk_time, total_work, );
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(extern Cilk_time, last_cp_time,);)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(extern Cilk_time, cp_hack, );)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(extern Cilk_time, work_hack, );)

DECLARE_PRIVATE_VARIABLE(extern Cilk_time, user_work, );
DECLARE_PRIVATE_VARIABLE(extern Cilk_time, user_critical_path ,);
#define Cilk_user_work USE_PRIVATE(user_work)
#define Cilk_user_critical_path USE_PRIVATE(user_critical_path)

/******************************************************
 *  structs containing shared/private variables, etc.
 ******************************************************/
#include <cilk-segments.h>

/***********************************************************\
 * Hacks
\***********************************************************/
#include "hacks.h"

/***********************************************************\
 * Hooks
\***********************************************************/
typedef void (*HookT)(void);

typedef struct hook {
     HookT fn;
     struct hook *next;
} HookList;

extern void Cilk_add_hook(HookList **listp, HookT fn);
extern void Cilk_run_hooks(HookList *list);

#define NULL_HOOK ((HookList *) 0)

/*
 * various types of hooks 
 * This list is incomplete and should be updated as needed
 */
extern HookList *Cilk_init_global_hooks;
extern HookList *Cilk_init_per_process_hooks;

/***********************************************************\
 * Critical path elapsed time calculator
\***********************************************************/

#if CILK_CRITICAL_PATH

DECLARE_PRIVATE_VARIABLE(extern Cilk_time, last_cp_time,);

static inline Cilk_time Cilk_get_elapsed_time(void)
{
     Cilk_time then = USE_PRIVATE(last_cp_time);
     Cilk_time now = Cilk_get_time();

     CILK_ASSERT(now >= then);

     USE_PRIVATE(last_cp_time) = now;
     return now - then;
}

#endif				/* CILK_CRITICAL_PATH */

/* internal malloc stuff, used by cilk2c's output */

/* statistics for Cilk_internal_malloc, used mainly for debugging */
struct Cilk_im_stats {
     int used;         /* bytes used; the sum must be 0 at the end */
     int wasted;       /* wasted because of the way internal_malloc works */
     int allocated;    /* allocated from the central pool */
     CILK_CACHE_LINE_PAD;   /* pad to cache line to avoid false sharing */
};

/*
 * this is written so that it can be inlined and partially
 * evaluated when size is a constant 
 */
#define CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, n) \
     if (size <= n) return n;

static inline int Cilk_internal_malloc_canonicalize(size_t size)
{
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 16);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 32);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 64);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 128);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 256);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 512);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 1024);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 2048);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 4096);
     return -1;  /* keep gcc happy */
}

#define CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, n, bucket) \
     if (size <= n) return bucket;

static inline int Cilk_internal_malloc_size_to_bucket(size_t size)
{
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 16, 0);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 32, 1);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 64, 2);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 128, 3);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 256, 4);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 512, 5);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 1024, 6);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 2048, 7);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 4096, 8);
     return -1;  /* keep gcc happy */
}

DECLARE_PRIVATE_VARIABLE(extern void *, im_free_lists, [CILK_INTERNAL_MALLOC_FREELISTS]);
DECLARE_SHARED_RO_VARIABLE(extern struct Cilk_im_stats *, im_info, );

static inline void *Cilk_internal_malloc_fast(size_t size)
{
     int bucket;
     void *mem;

     if (size > CILK_INTERNAL_MALLOC_THRESHOLD) 
	  return Cilk_internal_malloc(size);

     bucket = Cilk_internal_malloc_size_to_bucket(size);

     /* look at the free list */
     if ((mem = USE_PRIVATE(im_free_lists)[bucket])) {
	  WHEN_CILK_DEBUG({
	       USE_SHARED_RO(im_info)[Self].used += size;
	  })
	  USE_PRIVATE(im_free_lists)[bucket] = *((void **) mem);
     } else {
	  mem = Cilk_internal_malloc(size);
     }

     return mem;
}

static inline void Cilk_internal_free_fast(void *p, size_t size)
{
     int bucket;

     if (size > CILK_INTERNAL_MALLOC_THRESHOLD) {
	  Cilk_internal_free(p, size);
	  return;
     }

     bucket = Cilk_internal_malloc_size_to_bucket(size);

     WHEN_CILK_DEBUG({
	  USE_SHARED_RO(im_info)[Self].used -= size;
     })

     *((void **)p) = USE_PRIVATE(im_free_lists)[bucket];
     USE_PRIVATE(im_free_lists)[bucket] = p;
}

static inline void Cilk_destroy_frame(StackFrame *f, size_t size)
{
     WHEN_CILK_ALLOCA({
	  struct alloca_header *p;
	  struct alloca_header *q;
	  for (p = f->alloca_h; p; p = q) {
	       q = p->next;
	       Cilk_internal_free(p, p->size);
	  }
     })

     Cilk_internal_free_fast(f, size);
}

static inline void *Cilk_create_frame(size_t size, ProcInfo *sig)
{
     StackFrame *f = Cilk_internal_malloc_fast(size);
     f->sig = sig;
     WHEN_CILK_ALLOCA(f->alloca_h = (struct alloca_header *)0);
     WHEN_CILK_DEBUG(f->magic = CILK_STACKFRAME_MAGIC);

     return (void *)f;
}

/***********************************************************
 *  Cilk2C-only stuff
 ***********************************************************/
#ifdef __CILK2C__
extern cilk int main();
#endif

#endif				/* _CILK_H */
